/**
 * Copyright (C) 2011 Angelo Zerr <angelo.zerr@gmail.com> and Pascal Leclercq <pascal.leclercq@gmail.com>
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free  of charge, to any person obtaining
 * a  copy  of this  software  and  associated  documentation files  (the
 * "Software"), to  deal in  the Software without  restriction, including
 * without limitation  the rights to  use, copy, modify,  merge, publish,
 * distribute,  sublicense, and/or sell  copies of  the Software,  and to
 * permit persons to whom the Software  is furnished to do so, subject to
 * the following conditions:
 *
 * The  above  copyright  notice  and  this permission  notice  shall  be
 * included in all copies or substantial portions of the Software.
 *
 * THE  SOFTWARE IS  PROVIDED  "AS  IS", WITHOUT  WARRANTY  OF ANY  KIND,
 * EXPRESS OR  IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE  WARRANTIES OF
 * MERCHANTABILITY,    FITNESS    FOR    A   PARTICULAR    PURPOSE    AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE,  ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package fr.opensagres.xdocreport.template;

import static fr.opensagres.xdocreport.template.utils.TemplateUtils.getCachedTemplateName;

import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;

import org.apache.commons.io.IOUtils;

import fr.opensagres.xdocreport.core.XDocReportException;
import fr.opensagres.xdocreport.core.io.IEntryReaderProvider;
import fr.opensagres.xdocreport.core.io.IEntryWriterProvider;
import fr.opensagres.xdocreport.core.io.MultiWriter;
import fr.opensagres.xdocreport.core.utils.StringUtils;
import fr.opensagres.xdocreport.template.cache.ITemplateCacheInfoProvider;
import fr.opensagres.xdocreport.template.config.ITemplateEngineConfiguration;
import fr.opensagres.xdocreport.template.debug.ITemplateEngineDebugger;

public abstract class AbstractTemplateEngine implements ITemplateEngine {

	private ITemplateCacheInfoProvider templateCacheInfoProvider;
	private ITemplateEngineConfiguration configuration;

	public ITemplateCacheInfoProvider getTemplateCacheInfoProvider() {
		return templateCacheInfoProvider;
	}

	public void setTemplateCacheInfoProvider(
			ITemplateCacheInfoProvider templateCacheInfoProvider) {
		this.templateCacheInfoProvider = templateCacheInfoProvider;
	}

	public ITemplateEngineConfiguration getConfiguration() {
		return configuration;
	}

	public void setConfiguration(ITemplateEngineConfiguration configuration) {
		this.configuration = configuration;
	}

	public void process(String reportId, String entryName,
			IEntryReaderProvider readerProvider,
			IEntryWriterProvider writerProvider, IContext context,
			ITemplateEngineDebugger debugger) throws XDocReportException,
			IOException {

		// Get writer of the entry to merge Java model with template engine
		Writer writer = writerProvider.getEntryWriter(entryName);
		process(reportId, entryName, readerProvider, writer, context, debugger);
	}

	public void process(String reportId, String entryName,
			IEntryReaderProvider readerProvider, Writer writer,
			IContext context, ITemplateEngineDebugger debugger)
			throws XDocReportException, IOException {
		boolean useTemplateCache = isUseTemplateCache(reportId);
		// 1) Start process template engine
		long startTime = -1;
		if (isDebuggable(debugger)) {
			// Debug start preprocess
			startTime = System.currentTimeMillis();
			debugger.startProcessTemplateEngine(getId(), entryName,
					useTemplateCache);
		}

		Reader reader = null;
		try {
			writer = getWriter(writer, debugger);
			if (useTemplateCache) {
				// cache template is used, process it
				processWithCache(getCachedTemplateName(reportId, entryName),
						context, writer, debugger);
			} else {
				// No cache template is used, get the reader from the entry
				reader = readerProvider.getEntryReader(entryName);
				processNoCache(entryName, context, reader, writer, debugger);
			}
			if (isDebuggable(debugger)) {
				if (debugger.isDisplayResult()) {
					// Debug display result of preprocess
					debugger.resultOfProcessTemplateEngine(getId(), entryName,
							((MultiWriter) writer).getWriter(1).toString());
				}
				// Debug end preprocess
				debugger.endProcessTemplateEngine(getId(), entryName,
						System.currentTimeMillis() - startTime, null);
			}

		} catch (Throwable e) {
			if (isDebuggable(debugger)) {
				// Debug end processs with error
				debugger.endProcessTemplateEngine(getId(), entryName,
						System.currentTimeMillis() - startTime, e);
			}
			if (e instanceof RuntimeException) {
				throw (RuntimeException) e;
			}
			if (e instanceof IOException) {
				throw (IOException) e;
			}
			if (e instanceof XDocReportException) {
				throw (XDocReportException) e;
			}
			throw new XDocReportException(e);
		} finally {
			if (reader != null) {
				IOUtils.closeQuietly(reader);
			}
			if (writer != null) {
				IOUtils.closeQuietly(writer);
			}
		}
	}

	public void process(String entryName, IContext context, Reader reader,
			Writer writer, ITemplateEngineDebugger debugger)
			throws XDocReportException, IOException {
		try {
			processNoCache(entryName, context, reader, writer, debugger);
		} finally {
			if (reader != null) {
				IOUtils.closeQuietly(reader);
			}
			if (writer != null) {
				IOUtils.closeQuietly(writer);
			}
		}

	}

	public void extractFields(IEntryReaderProvider readerProvider,
			String entryName, FieldsExtractor extractor)
			throws XDocReportException {
		Reader reader = readerProvider.getEntryReader(entryName);
		extractFields(reader, entryName, extractor);
	}

	protected boolean isDebuggable(ITemplateEngineDebugger debugger) {
		return debugger != null;
	}

	private Writer getWriter(Writer writer, ITemplateEngineDebugger debugger) {
		if (isDebuggable(debugger) && debugger.isDisplayResult()) {
			return new MultiWriter(writer, new StringWriter());
		}
		return writer;
	}

	protected boolean isUseTemplateCache(String reportId) {
		return StringUtils.isNotEmpty(reportId)
				&& getTemplateCacheInfoProvider() != null
				&& getTemplateCacheInfoProvider().existsReport(reportId);

	}

	protected abstract void processWithCache(String templateName,
			IContext context, Writer writer, ITemplateEngineDebugger debugger)
			throws XDocReportException, IOException;

	protected abstract void processNoCache(String entryName, IContext context,
			Reader reader, Writer writer, ITemplateEngineDebugger debugger)
			throws XDocReportException, IOException;
}
