/**
 * Copyright (C) 2011 Angelo Zerr <angelo.zerr@gmail.com> and Pascal Leclercq <pascal.leclercq@gmail.com>
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free  of charge, to any person obtaining
 * a  copy  of this  software  and  associated  documentation files  (the
 * "Software"), to  deal in  the Software without  restriction, including
 * without limitation  the rights to  use, copy, modify,  merge, publish,
 * distribute,  sublicense, and/or sell  copies of  the Software,  and to
 * permit persons to whom the Software  is furnished to do so, subject to
 * the following conditions:
 *
 * The  above  copyright  notice  and  this permission  notice  shall  be
 * included in all copies or substantial portions of the Software.
 *
 * THE  SOFTWARE IS  PROVIDED  "AS  IS", WITHOUT  WARRANTY  OF ANY  KIND,
 * EXPRESS OR  IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE  WARRANTIES OF
 * MERCHANTABILITY,    FITNESS    FOR    A   PARTICULAR    PURPOSE    AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE,  ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package fr.opensagres.xdocreport.template.registry;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import fr.opensagres.xdocreport.core.document.DocumentKind;
import fr.opensagres.xdocreport.core.registry.AbstractRegistry;
import fr.opensagres.xdocreport.core.utils.StringUtils;
import fr.opensagres.xdocreport.template.ITemplateEngine;
import fr.opensagres.xdocreport.template.TemplateEngineKind;
import fr.opensagres.xdocreport.template.discovery.ITemplateEngineDiscovery;
import fr.opensagres.xdocreport.template.discovery.ITemplateEngineInitializerDiscovery;

/**
 * Template engien registry which store instance of {@link ITemplateEngine} for
 * template engine kind (Freemarker, Velocity) and document kind (odt, docx..)
 * 
 */
public class TemplateEngineRegistry extends AbstractRegistry implements
		Serializable {

	private static final long serialVersionUID = -7686229087661483932L;
	private static final String TEMPLATE_DISCOVERY_PROPERTIES = "template-discovery.properties";
	private static final TemplateEngineRegistry INSTANCE = new TemplateEngineRegistry();

	private Map<String, ITemplateEngine> templateEnginesCache = new HashMap<String, ITemplateEngine>();
	private Map<String, ITemplateEngineInitializerDiscovery> templateEnginesInitializerDiscoveryCache = new HashMap<String, ITemplateEngineInitializerDiscovery>();
	private Map<String, ITemplateEngineDiscovery> templateEnginesDiscoveryCache = new HashMap<String, ITemplateEngineDiscovery>();
	private ITemplateEngine defaultTemplateEngine;
	private Collection<String> templateEngineKinds = new ArrayList<String>();

	public static TemplateEngineRegistry getRegistry() {
		return INSTANCE;
	}

	public TemplateEngineRegistry() {
		super(TEMPLATE_DISCOVERY_PROPERTIES, TemplateEngineRegistry.class
				.getSimpleName());
	}

	public ITemplateEngine getTemplateEngine(
			TemplateEngineKind templateEngineKind, DocumentKind documentKind) {
		return getTemplateEngine(templateEngineKind.name(),
				documentKind != null ? documentKind.name() : null);
	}

	public ITemplateEngine getTemplateEngine(String templateEngineKind,
			String documentKind) {
		initializeIfNeeded();
		String key = getKey(templateEngineKind, documentKind);
		return templateEnginesCache.get(key);
	}

	private String getKey(String templateEngineKind, String documentKind) {
		StringBuilder key = new StringBuilder(templateEngineKind);
		if (StringUtils.isNotEmpty(documentKind)) {
			key.append('_');
			key.append(documentKind);
		}
		return key.toString();
	}

	public void register(ITemplateEngine templateEngine) {
		register(templateEngine, (String) null);
	}

	public void register(ITemplateEngine templateEngine,
			DocumentKind documentKind) {
		register(templateEngine, documentKind.name());
	}

	public void register(ITemplateEngine templateEngine, String documentKind) {
		String templateEngineKind = templateEngine.getKind();
		String key = getKey(templateEngineKind, documentKind);
		templateEnginesCache.put(key, templateEngine);
		if (defaultTemplateEngine == null) {
			defaultTemplateEngine = templateEngine;
		}
		if (!templateEngineKinds.contains(templateEngineKind)) {
			templateEngineKinds.add(templateEngineKind);
		}
	}

	@Override
	protected boolean registerInstance(Object instance) {
		if (instance instanceof ITemplateEngineDiscovery) {
			ITemplateEngineDiscovery discovery = (ITemplateEngineDiscovery) instance;
			templateEnginesDiscoveryCache.put(discovery.getId(), discovery);
			ITemplateEngine templateEngine = discovery.createTemplateEngine();
			register(templateEngine);
			return true;
		}
		if (instance instanceof ITemplateEngineInitializerDiscovery) {
			ITemplateEngineInitializerDiscovery discovery = (ITemplateEngineInitializerDiscovery) instance;
			templateEnginesInitializerDiscoveryCache.put(discovery.getId(),
					discovery);
		}
		return false;
	}

	@Override
	protected void onEndInitialization() {
		// When registry end initialisation, loop for each initializer to
		// initialize each template engine.
		Collection<ITemplateEngineInitializerDiscovery> discoveriesForWholeTemplateEngine = new ArrayList<ITemplateEngineInitializerDiscovery>();
		Collection<ITemplateEngineInitializerDiscovery> discoveries = templateEnginesInitializerDiscoveryCache
				.values();
		String documentKind = null;
		for (ITemplateEngineInitializerDiscovery discovery : discoveries) {
			documentKind = discovery.getDocumentKind();
			if (StringUtils.isEmpty(documentKind)) {
				discoveriesForWholeTemplateEngine.add(discovery);
			} else {
				Set<Entry<String, ITemplateEngineDiscovery>> entries = templateEnginesDiscoveryCache
						.entrySet();
				for (Entry<String, ITemplateEngineDiscovery> entry : entries) {
					String templateEngineKind = entry.getKey();
					ITemplateEngineDiscovery templateEngineDiscovery = entry
							.getValue();

					ITemplateEngine templateEngine = getTemplateEngine(
							templateEngineKind, documentKind);
					if (templateEngine == null) {
						// Create template engine for template/document kind
						templateEngine = templateEngineDiscovery
								.createTemplateEngine();
						register(templateEngine, documentKind);
					}
					discovery.initialize(templateEngine);
				}
			}
		}

		for (ITemplateEngineInitializerDiscovery discovery : discoveriesForWholeTemplateEngine) {
			Collection<ITemplateEngine> templateEngines = getTemplateEngines();
			for (ITemplateEngine templateEngine : templateEngines) {
				discovery.initialize(templateEngine);
			}
		}

	}

	public boolean exists(TemplateEngineKind templateEngineKind,
			DocumentKind documentKind) {
		return exists(templateEngineKind.name(),
				documentKind != null ? documentKind.name() : null);
	}

	public boolean exists(String templateEngineKind, String documentKind) {
		initializeIfNeeded();
		String key = getKey(templateEngineKind, documentKind);
		return templateEnginesCache.containsKey(key);
	}

	public boolean isDefault(ITemplateEngine templateEngine) {
		initializeIfNeeded();
		if (templateEngine == null) {
			return false;
		}
		return templateEngine.equals(defaultTemplateEngine);
	}

	public ITemplateEngine getDefaultTemplateEngine() {
		initializeIfNeeded();
		return defaultTemplateEngine;
	}

	public void setDefaultTemplateEngine(ITemplateEngine defaultTemplateEngine) {
		this.defaultTemplateEngine = defaultTemplateEngine;
	}

	public ITemplateEngine getTemplateEngine(String templateEngineId) {
		initializeIfNeeded();
		String templateEngineKind = templateEngineId;
		String documentKind = null;
		int index = templateEngineId.lastIndexOf('_');
		if (index != -1) {
			templateEngineKind = templateEngineId.substring(0, index);
			documentKind = templateEngineId.substring(index + 1,
					templateEngineId.length());
		}
		return getTemplateEngine(templateEngineKind, documentKind);
	}

	public Collection<ITemplateEngine> getTemplateEngines() {
		initializeIfNeeded();
		return templateEnginesCache.values();
	}

	public Collection<String> getTemplateEngineKinds() {
		initializeIfNeeded();
		return templateEngineKinds;
	}

	@Override
	protected void doDispose() {
		this.templateEnginesCache.clear();
		this.templateEnginesInitializerDiscoveryCache.clear();
		this.templateEnginesDiscoveryCache.clear();
		this.defaultTemplateEngine = null;
		this.templateEngineKinds.clear();

	}
}
